{$MODE objfpc}{$H+}

unit woprspell;

interface
uses Classes, dynlibs, SysUtils;

const MAX_NAME = 20;
const MAX_CHARS = 100;

type

  PHyphenTrans = ^HyphenTrans;
  HyphenTrans = record
    ch: Char;
    new_state: Integer;
  end;

  PHyphenState = ^HyphenState;
  HyphenState = record
    match: PChar;
    repl: PChar;
    replindex: Char;
    replcut: Char;
    fallback_state: Integer;
    num_trans: Integer;
    trans: PHyphenTrans;
  end;

  PHyphenDict = ^HyphenDict;
  HyphenDict = record
    lhmin: char;
    rhmin: char;
    clhmin: char;
    crhmin: char;
    num_states: integer;
    cset: array[0..MAX_NAME - 1] of Char;
    utf8: integer;
    states: PHyphenState;
    nextlevel: PHyphenDict;
  end;

  Thnj_hyphen_load = function(const fn: PChar): PHyphenDict cdecl;
  Thnj_hyphen_free = procedure(var dict: HyphenDict) cdecl;
  Thnj_hyphen_hyphenate = function(var dict: HyphenDict; const word: PChar; word_size: Integer; hyphens: PChar): Integer cdecl;
  Thnj_hyphen_hyphenate2 = function(var dict: HyphenDict; const word: PChar; word_size: Integer; hyphens: PChar; hyphenated_word: PChar;
    var rep: PPChar; var pos: Pointer; var cut: Pointer): Integer cdecl;

  THunspell_create = function(aff_file: PChar; dict_file: PChar): Pointer; cdecl;
  THunspell_destroy = procedure(spell: Pointer); cdecl;
  THunspell_spell = function(spell: Pointer; word: PChar): Boolean; cdecl;
  THunspell_suggest = function(spell: Pointer; var slst: PPChar; word: PChar): Integer; cdecl;
  THunspell_analyze = function(spell: Pointer; var slst: PPChar; word: PChar): Integer; cdecl;
  THunspell_stem = function(spell: Pointer; var slst: PPChar; word: PChar): Integer; cdecl;
  THunspell_free_list = procedure(spell: Pointer; var slst: PPChar; n: integer); cdecl;
  THunspell_get_dic_encoding = function(spell: Pointer): PChar; cdecl;
  THunspell_add = function(spell: Pointer; word: PChar): Integer; cdecl;
  THunspell_remove = function(spell: Pointer; word: PChar): Integer; cdecl;

  { THyphen }

  THyphen = class
  private
    Pdict: PHyphenDict;
  public
    constructor Create(DictionaryPath: string; LibraryName: string = '');
    destructor Destroy; override;
    function Hyphenate(Word: string): string;
    procedure Hyphenate(Word: string; BreakAfterList: TList);
  end;

  { THunspell }

  THunspell = class
  private
    fAspell: Pointer;
    fDicUTF8: Boolean;
  public
    constructor Create(AffFile, DictFile: string; LibraryName: string = '');
    destructor Destroy; override;
    function Spell(Word: string): boolean;
    procedure Suggest(Word: string; List: TStrings);
    procedure Add(Word: string);
    procedure Remove(Word: string);
  end;

var hnj_hyphen_load: Thnj_hyphen_load;
var hnj_hyphen_free: Thnj_hyphen_free;
var hnj_hyphen_hyphenate: Thnj_hyphen_hyphenate;
var hnj_hyphen_hyphenate2: Thnj_hyphen_hyphenate2;

var Hunspell_create: THunspell_create;
var Hunspell_destroy: THunspell_destroy;
var Hunspell_spell: Thunspell_spell;
var Hunspell_suggest: Thunspell_suggest;
var Hunspell_analyze: Thunspell_analyze;
var Hunspell_stem: Thunspell_stem;
var Hunspell_get_dic_encoding: Thunspell_get_dic_encoding;
var Hunspell_add: THunspell_add;
var Hunspell_free_list: THunspell_free_list;
var Hunspell_remove: THunspell_remove;

var HunLibLoaded: Boolean = False;
var HunLibHandle: THandle;
var HypLibLoaded: Boolean = False;
var HypLibHandle: THandle;

function LoadLibHunspell(LibraryName: string): Boolean;
function LoadLibHyphen(LibraryName: string): Boolean;

implementation

function LoadLibHunspell(libraryName: string): Boolean;
begin
  if libraryName = '' then
    {$IFDEF windows}libraryName := 'hunspelldll.dll'; {$ENDIF}
  {$IFDEF linux}libraryName := 'libhunspell-1.2.so'; {$ENDIF}

  Result := HunLibLoaded;
  if Result then
    exit;

  HunLibHandle := LoadLibrary(PAnsiChar(libraryName));
  if HunLibHandle <> 0 then begin
    Result := True;

    Hunspell_create := THunspell_create(GetProcAddress(HunLibHandle, 'Hunspell_create'));
    if not Assigned(hunspell_create) then Result := False;
    Hunspell_destroy := Thunspell_destroy(GetProcAddress(HunLibHandle, 'Hunspell_destroy'));
    if not Assigned(hunspell_destroy) then Result := False;
    Hunspell_spell := THunspell_spell(GetProcAddress(HunLibHandle, 'Hunspell_spell'));
    if not Assigned(hunspell_spell) then Result := False;
    Hunspell_suggest := THunspell_suggest(GetProcAddress(HunLibHandle, 'Hunspell_suggest'));
    if not Assigned(hunspell_suggest) then Result := False;
    Hunspell_analyze := THunspell_analyze(GetProcAddress(HunLibHandle, 'Hunspell_analyze'));
    if not Assigned(hunspell_analyze) then Result := False;
    Hunspell_stem := THunspell_stem(GetProcAddress(HunLibHandle, 'Hunspell_stem'));
    if not Assigned(hunspell_stem) then Result := False;
    Hunspell_get_dic_encoding := THunspell_get_dic_encoding(GetProcAddress(HunLibHandle, 'Hunspell_get_dic_encoding'));
    if not Assigned(hunspell_get_dic_encoding) then Result := False;
    Hunspell_free_list := THunspell_free_list(GetProcAddress(HunLibHandle, 'Hunspell_free_list'));
    if not Assigned(Hunspell_free_list) then Result := False;
    Hunspell_add := THunspell_add(GetProcAddress(HunLibHandle, 'Hunspell_add'));
    if not Assigned(Hunspell_add) then Result := False;
    Hunspell_remove := THunspell_remove(GetProcAddress(HunLibHandle, 'Hunspell_remove'));
    if not Assigned(Hunspell_remove) then Result := False;

    HunLibLoaded := Result;
  end;
end;

{ THunspell }

constructor THunspell.Create(AffFile, DictFile: string; LibraryName: string);
begin
  if LoadLibHunspell(LibraryName) then
    fAspell := hunspell_create(PChar(AffFile), PChar(DictFile)) else
    raise Exception.Create('Could not load the Hunspell library');
  fDicUTF8 := Lowercase(Hunspell_get_dic_encoding(fASpell)) = 'utf-8';
  //writeln(fDicUTF8);
end;

destructor THunspell.Destroy;
begin
  if (HunLibHandle <> 0) and HunLibLoaded then
  begin
  if fAspell<>nil then hunspell_destroy(fAspell);
  fAspell:=nil;
  if HunLibHandle <> 0 then FreeLibrary(HunLibHandle);
  HunLibLoaded := false;
  end;
  inherited Destroy;
end;

procedure LongToWideBuf(const Buffer: Pointer; const BufferSize: Integer;
  const DestBuffer: Pointer);
var I: Integer;
  S: Pointer;
  D: Pointer;
  LW: LongWord;
begin
  if BufferSize <= 0 then
    exit;
  S := Buffer;
  D := DestBuffer;
  for I := 1 to BufferSize div 4 do
  begin
    LW := PLongWord(S)^;
    Inc(PLongWord(S));
    PLongWord(D)^ := (LW and $FF) or ((LW and $FF00) shl 8);
    Inc(PLongWord(D));
    LW := LW shr 16;
    PLongWord(D)^ := (LW and $FF) or ((LW and $FF00) shl 8);
    Inc(PLongWord(D));
  end;
  for I := 1 to BufferSize mod 4 do
  begin
    PWord(D)^ := PByte(S)^;
    Inc(PByte(S));
    Inc(PWord(D));
  end;
end;

function LongToWideString(const S: AnsiString): WideString;
var L: Integer;
begin
  L := Length(S);
  SetLength(Result, L);
  if L = 0 then
    exit;
  LongToWideBuf(Pointer(S), L, Pointer(Result));
end;

function PWideToLongString(const P: PWideChar; const Len: Integer): AnsiString;
var I: Integer;
  S: PWideChar;
  Res: PAnsiChar;
  WC: WideChar;
begin
  if Len <= 0 then
  begin
    Result := '';
    exit;
  end;
  SetLength(Result, Len);
  S := P;
  Res := Pointer(Result);
  for I := 1 to Len do
  begin
    WC := S^;
    Res^ := AnsiChar(Byte(WC));
    Inc(S);
    Inc(Res);
  end;
end;

function WideToLongString(const S: WideString): AnsiString;
begin
  Result := PWideToLongString(Pointer(S), Length(S));
end;

function THunspell.Spell(Word: string): boolean;
begin
  if fDicUTF8 then
    Result := hunspell_spell(fAspell, PChar(Word)) else
    Result := hunspell_spell(fAspell, PChar(WideToLongString(UTF8Decode(Word))));
end;

procedure THunspell.Suggest(Word: string; List: TStrings);
var i, len: Integer;
var owds, wrds: PPChar;
begin
  List.clear;
  try
    if fDicUTF8 then
    begin
      len := hunspell_suggest(fAspell, owds, PChar(Word));
      wrds := owds;
      for i := 1 to len do begin
        List.Add(wrds^);
        Inc(PtrInt(wrds), sizeOf(Pointer));
      end;
    end else
    begin
      len := hunspell_suggest(fAspell, owds, PChar(WideToLongString(UTF8Decode(Word))));
      wrds := owds;
      for i := 1 to len do begin
        List.Add(UTF8Encode(LongToWideString(wrds^)));
        Inc(PtrInt(wrds), sizeOf(Pointer));
      end;
    end;
  finally
    Hunspell_free_list(fAspell, owds, len);
  end;
end;

procedure THunspell.Add(Word: string);
begin
  Hunspell_add(fASpell, Pchar(Word));
end;

procedure THunspell.Remove(Word: string);
begin
  Hunspell_remove(fASpell, Pchar(Word));
end;

function LoadLibHyphen(libraryName: string): Boolean;
begin
  if libraryName = '' then
    {$IFDEF windows}libraryName := 'hyphen.dll'; {$ENDIF}
  {$IFDEF linux}libraryName := 'libhyphen.so'; {$ENDIF}

  Result := HypLibLoaded;
  if Result then
    exit;

  HypLibHandle := LoadLibrary(PAnsiChar(libraryName));
  if HypLibHandle <> 0 then begin
    Result := True;

    hnj_hyphen_load := Thnj_hyphen_load(GetProcAddress(HypLibHandle, 'hnj_hyphen_load'));
    if not Assigned(hnj_hyphen_load) then Result := False;
    hnj_hyphen_free := Thnj_hyphen_free(GetProcAddress(HypLibHandle, 'hnj_hyphen_free'));
    if not Assigned(hnj_hyphen_free) then Result := False;
    hnj_hyphen_hyphenate := Thnj_hyphen_hyphenate(GetProcAddress(HypLibHandle, 'hnj_hyphen_hyphenate'));
    if not Assigned(hnj_hyphen_hyphenate) then Result := False;
    hnj_hyphen_hyphenate2 := Thnj_hyphen_hyphenate2(GetProcAddress(HypLibHandle, 'hnj_hyphen_hyphenate2'));
    if not Assigned(hnj_hyphen_hyphenate2) then Result := False;

    HypLibLoaded := Result;
  end;
end;

{ THyphen }

constructor THyphen.Create(DictionaryPath: string; LibraryName: string);
begin
  if LoadLibHyphen(libraryName) then
    pdict := hnj_hyphen_load(Pchar(DictionaryPath)) else
    raise Exception.Create('Could not load the Hyphen library');
  //   writeln(pdict^.utf8);
end;

destructor THyphen.Destroy;
begin
  if (HypLibHandle <> 0) and HypLibLoaded then
  begin
  if pdict<>nil then hnj_hyphen_free(pdict^);
  pdict:=nil;
  if HypLibHandle <> 0 then FreeLibrary(HypLibHandle);
  HypLibLoaded := false;
  end;
  inherited Destroy;
end;


function THyphen.Hyphenate(Word: string): string;
var hyphens: PChar;
  hyphenated_word: PChar;
  rep: PPChar;
  pos: Pchar;
  cut: PChar;
  len: integer;
begin
  if not assigned(pdict) then raise Exception.Create('hyphen dictionary not loaded');
  len := Length(Word);
  hyphens := StrAlloc(Len + 5);
  hyphenated_word := StrAlloc(len * 2);
  rep := nil;
  pos := nil;
  cut := nil;
  try
    if hnj_hyphen_hyphenate2(pdict^, Pchar(Word), Len, hyphens, hyphenated_word, rep, pos, cut) = 0 then
    begin
      result := hyphenated_word;
    end else result := Word;
  finally
    StrDispose(hyphens);
    StrDispose(hyphenated_word);
  end;
end;

procedure THyphen.Hyphenate(Word: string; BreakAfterList: TList);
var res: string;
  i, count: integer;
begin
  BreakAfterList.clear;
  res := Hyphenate(Word);
  count := 0;
  for i := 1 to length(res) do
  begin
    if res[i] = '=' then
    begin
      BreakAfterList.Add(Pointer(count));
      dec(count);
    end;
    inc(count);
  end;
end;

end.
